

   // We assume that theta does not imply a blocking when called,
   // and that g does not contain an empty clause. 

   bool notlearning:forward( gcsp& g, 
                             subststack< size_t > & theta, size_t s,
                             std::vector< refinement > & stack )
   {
      while( s < theta. size( ))
      {
         variable v = theta[s]. first;

         const auto& occ = g. inclauses(v);
         for( auto c : occ )
         {
            // It is not a for loop, because we either delete or 
            // move to the next.

            auto original_state = g. clauses[c]. current;

            size_t i = g. clauses[c]. current. s0;  
            while( i != g. clauses[c]. current. s1 ) 
            {
               if( consistent( theta, g. clauses[c], i ) &&
                   ! g.impliesblocked( theta, g. clauses[c], i ))
               {
                  ++ i;
               }
               else
               {
                  g. clauses[c]. disable(i); 
               }
            }

            if( original_state. size( ) != g. clauses[c]. current. size( ))
            {
               if( g. clauses[c]. current. size( ) == 0 )
               {
                  g. clauses[c]. current = original_state; 
                  return false;
               } 

               stack. push_back( { c, {c}, original_state, 
                                   g. clauses[c]. current, theta. size( ) } ); 

               if( g. clauses[c]. current. size( ) == 1 )
               {
                  match( theta, g.clauses[c], g. clauses[c]. current. s0 );
               }
            }
         }
         ++ s; 
      }
      return true;
   }


   // We assume that forward reasoning has been exhaustively applied.
   // If stopfirst == true, we stop on the first solution, and this solution
   // remains in theta.

   void notlearning::solve( gcsp& g, 
                            subststack< size_t > & theta, 
                            std::vector< refinement > & stack, 
                            size_t& nrsolutions )
   {
      size_t picked = pick(g); 

      if( picked == g. clauses. size( ))
      {
         printsolution( g, theta, nrsolutions ); 
      }
      else
      {
         auto restore = g. clauses[ picked ]. current;
         for( auto choice = restore. first(1); 
                   choice. size( );
                   choice. next( restore )) 
         {
            size_t s = theta. size( );
            size_t k = stack. size( ); 

            if( !match( theta, g. clauses[ picked ], choice. s0 ))
            {
               std::cout << "theta " << theta << "\n";
               throw std::runtime_error( "choice was unmatchable!" );
            }

            if( !g. impliesblocked( theta, s ))
            {
               g. clauses[ picked ]. current = choice; 

               stack. push_back( { picked, {}, restore, choice, s } );

               if( forward( g, theta, s, stack ))
                  solve( g, theta, stack, nrsolutions ); 
            
               while( stack. size( ) > k )
               {
                  g. clauses[ stack. back( ). cl ]. current
                         = stack. back( ). previous; 
                  stack. pop_back( ); 
               }
            }
            theta. restore(s);
         }
      }
   }



