
// Written by Hans de Nivelle, May 2016.

#include "preprocessor.h"
#include "resolution.h"


namespace geometric
{
namespace constraints
{

   namespace 
   {

      struct todo 
      {
         std::deque< size_t > deq;   // First time that I use dequeue. 

         todo( ) { } 

         todo( std::initializer_list< size_t > init )
         {
            for( auto i : init )
               insert(i);
         }

      
         void insert( size_t k )
         {
            for( auto i : deq )
               if( i == k ) return;
            deq. push_back(k);
         }


         // If you are certain that k does not occur:
 
         void insert_new( size_t k )
            { deq. push_back(k); }


         size_t select_remove( )   
         {
            size_t k = deq. front( );
            deq. pop_front( );
            return k; 
         }

         size_t size( ) const { return deq. size( ); }
      };
   
      inline std::ostream& operator << ( std::ostream& out, const todo d )
      {
         out << '{'; 
         for( auto p = d. deq. begin( ); 
              p != d. deq. end( );
              ++ p )
         {
            if( p != d. deq. begin( ))
               out << ',';
            out << *p;
         }
         out << '}';
         return out;
      }

      // Filter clause cl against cl1:

      bool filter_into( subststack< size_t > & theta,
                        gcsp& g, size_t cl, size_t cl1, todo& unchecked )
      {
         size_t i = g. clauses[ cl ]. current. s0;
         while( i < g. clauses[ cl ]. current. s1 ) 
         {
            if( solvable( theta, g, g. clauses[ cl ], i, g. clauses[ cl1 ] ))
               ++ i;
            else
            {
               g. clauses[ cl ]. disable(i); 
               unchecked. insert( cl );
            }
         }

         if( g.clauses[cl]. current. size( ) == 1 )
            match( theta, g.clauses[cl], g.clauses[cl]. current. s0 );

         return g.clauses[cl]. current. size( ) != 0;
      }


      bool filterfrom( subststack< size_t > & theta,
                       gcsp& g, size_t from, todo& unchecked )
      {
         for( size_t v = 0; v < g. clauses[ from ]. nrvars( ); ++ v )
         {
            const auto& occ = g. inclauses( g. clauses[ from ]. var(v) );
            for( auto into : occ )
            {
               if( from != into )
               {
                  if( filter_into( theta, g, into, from, unchecked ) == 0 )
                     return false;
               }
            }
         }

         return true;
      }
   }


   bool preprocessor( gcsp& g, subststack< size_t > & theta, size_t local )
   {
      // Check for blockings with empty domain:

      for( size_t b = 0; b < g. blockings. size( ); ++ b )
      {
         if( g. blockings[b]. nrvars( ) == 0 &&
             g. blockings[b]. current. size( ))
         {
            return false;
         }
      }

      // Remove substlets that imply a blocking from the clauses: 

      for( size_t c = 0; c < g. clauses. size( ); ++ c )
      {
         size_t j = g. clauses[c]. current. s0;
         while( j < g. clauses[c]. current. s1 )
         {
            if( g. impliesblocked( theta, g. clauses[c], j ))
               g. clauses[c]. disable(j);
            else
               ++ j;  
         }

         if( g. clauses[c]. current. size( ) == 0 )
            return false;
               // We managed to make the clause empty, or it
               // was already empty. 

         if( g. clauses[c]. current. size( ) == 1 )
            match( theta, g.clauses[c], g.clauses[c]. current. s0 );
      }

      if( local >= 2 )
      { 
         todo unchecked;
      
         for( size_t c = 0; c < g. clauses. size( ); ++ c )
         {
            if( !filterfrom( theta, g, c, unchecked ))
               return false;
         }

         size_t cnt = 0;

         while( unchecked. size( ))
         {
            size_t from = unchecked. select_remove( );

            if( !filterfrom( theta, g, from, unchecked ))
               return false;

            ++ cnt;
         } 

         // std::cout << cnt << " passes were needed for 2-filtering.\n"; 
      }

      return true; 
   }


}}


