
// Written by Hans de Nivelle, April/May 2016.

#include "resolution.h"

namespace geometric
{
namespace constraints
{

   bool solvable( subststack< size_t > & theta,
                  const gcsp& g,
                  const gcsp::clause& cl, size_t i,
                  const gcsp::clause& cl1 )
   {
      const size_t s = theta. size( );

      if( match( theta, cl, i ) && !g.impliesblocked( theta, s ))
      {
         const size_t s1 = theta. size( );

         for( size_t i1 = cl1. current. s0; i1 != cl1. current. s1; ++ i1 )
         {
            if( match( theta, cl1, i1 ) && !g. impliesblocked( theta, s1 ))
            {
                theta. restore(s);
                return true;
            }
 
            theta. restore(s1);
         }
      }
     
      theta. restore(s); 
      return false; 
   }


   bool solvable( subststack< size_t > & theta,
                  const gcsp& g,
                  const gcsp::clause& cl, size_t i,
                  const gcsp::clause& cl1, const gcsp::clause& cl2 )
   {

      const size_t s = theta. size( );

      if( match( theta, cl, i ) && !g.impliesblocked( theta, s ))
      {
         const size_t s1 = theta. size( );

         for( size_t i1 = cl1. current. s0; i1 != cl1. current. s1; ++ i1 )
         {
            if( match( theta, cl1, i1 ) && !g. impliesblocked( theta, s1 ))
            {
               const size_t s2 = theta. size( );
           
               for( size_t i2 = cl2. current. s0; 
                           i2 != cl2. current. s1; 
                           ++ i2 )
               {
                  if( match( theta, cl2, i2 ) && 
                      !g. impliesblocked( theta, s2 ))
                  {
                     theta. restore(s);
                     return true;
                  }

                  theta. restore(s2);
               }
            }
 
            theta. restore(s1);
         }
      }
     
      theta. restore(s); 
      return false; 
   }


   namespace
   {

      bool implies( const gcsp& g,
                    subststack< size_t > & theta, 
                    const std::vector< const gcsp::clause* > & clauses,
                    size_t clausedepth, 
                    const std::vector< const lemma* > & lemmas,
                    size_t lemmadepth, 
                    const lemma& conclusion )
      {
         // std::cout << "Checking implication ";
         // std::cout << "at clauses[ " << clausedepth << " ] ";
         // std::cout << "lemmas[ " << lemmadepth << " ]\n";

         // std::cout << theta << "\n";

         if( istrue( theta, conclusion ))
         {
            // std::cout << "conclusion is true\n";
            return true;
         }

         size_t s = theta. size( );

         if( clausedepth < clauses. size( ))  
         {
            for( size_t c = clauses[ clausedepth ] -> initial. s0; 
                        c < clauses[ clausedepth ] -> initial. s1; 
                        ++ c ) 
            {
               // std::cout << "checking clause " << clausedepth;
               // std::cout << "/" << c << "\n";

               if( match( theta, *clauses[ clausedepth ], c ) &&
                   !g. impliesblocked( theta, s ))
               {
                  if( !implies( g, theta, 
                                clauses, clausedepth + 1,
                                lemmas, lemmadepth, conclusion ))
                  {
                     return false; 
                  }
               }

               theta. restore(s); 
            }
            return true; 
         }

         if( lemmadepth < lemmas. size( ))  
         {
            for( size_t c = 0; 
                        c < lemmas[ lemmadepth ] -> nrsubstlets( ); ++ c ) 
            {
               // std::cout << "checking lemma " << lemmadepth;
               // std::cout << "/" << c << "\n";

               if( match( theta, *lemmas[ lemmadepth ], c ) && 
                   !g. impliesblocked( theta, s )) 
               {
                  if( !implies( g, theta, 
                                clauses, clausedepth,
                                lemmas, lemmadepth + 1, conclusion )) 
                  {
                     return false; 
                  }
               }
               
               theta. restore(s);  
            }
            return true; 
         }

         return false; 
            // We have given conclusion all possible chances, and could not
            // make it true. 

      }
   }


   bool implies( const gcsp& g,
                 const std::vector< const gcsp::clause* > & clauses,
                 const std::vector< const lemma* > & lemmas,
                 const lemma& conclusion )
   {
#if 0
      std::cout << "Implies\n";
      for( const auto p : clauses )
         std::cout << *p << "\n";
      for( const auto p : lemmas )
         std::cout << *p << "\n";
      std::cout << " |- " << conclusion << "\n\n";
#endif
      subststack< size_t > theta;
      return implies( g, theta, clauses, 0, lemmas, 0, conclusion );
   }


   namespace
   {

      // True if cl[i] conflicts all substlets in lem[ confl ].

      bool conflictsallconfl( const gcsp::clause& cl, size_t i,
                              const lemma& lem, const confl_set& confl )
      {
         for( size_t j : confl )
         {
            if( consistent( lem, j, cl, i )) 
               return false;
         }
         return true;
      }


      void 
      addproducts( lemma& lem, 
            std::vector< std::pair< const gcsp::clause*, size_t >> & chosen,
            size_t level, 
            const std::vector< const gcsp::clause* > & reducers )
      {
         if( level == reducers. size( ))
         {
            lem. insert( chosen );
         }
         else
         {
            for( size_t i = reducers[ level ] -> current. s0; 
                        i != reducers[ level ] -> current. s1; 
                        ++ i )
            {
               chosen. push_back( { reducers[ level ], i } );  
               addproducts( lem, chosen, level + 1, reducers );
               chosen. pop_back( );
            }
         }
      }

   } 



   lemma resolvent( const gcsp::clause* mainparent, 
                    const std::vector< const lemma* > & conflictlemmas,
                    gcsp::clause::state before_reduction, 
                    const std::vector< const gcsp::clause* > & reducers )
   {
#if 0
      std::cout << "Constructing resolvent:\n";
      std::cout << "   main: " << *mainparent << "\n";
#endif

      if( !mainparent )
         throw std::runtime_error( "main parent must be present" );
 
      if( mainparent -> current. size( ) != conflictlemmas. size( ))
         throw std::runtime_error( "resolvent: lengths differ" );

      if( !mainparent -> current. included_in( before_reduction )) 
         throw std::runtime_error("current not included in before_reduction");

      lemma result; 

      // Go through the conflict lemmas and decide which substlets can 
      // be resolved away:

      std::vector< confl_set > conflicting = 
         std::vector< confl_set > ( conflictlemmas. size( ));

      for( size_t c = 0; c < conflictlemmas. size( ); ++ c )
      {
         if( conflictlemmas[c] )
         {
            for( size_t i = 0; i < conflictlemmas[c] -> nrsubstlets( ); ++ i )
            {
               if( !consistent( *conflictlemmas[c], i, 
                                *mainparent, mainparent -> current. s0 + c ))
               {
                  conflicting[c]. insert(i);
               }
            }
         }
      }

#if 0
      for( size_t c = 0; c < conflictlemmas. size( ); ++ c )
      {
         std::cout << "   conflictlemma: " << c << ": ";
         if( conflictlemmas[c] )
         {
            std::cout << *conflictlemmas[c] << "\n";
            std::cout << "      " << conflicting[c] << "\n";
         }
         else
            std::cout << "(nullptr)\n";
      }

      for( size_t r = 0; r < reducers. size( ); ++ r )
      {
         std::cout << "   reducers " << r << ": ";
         std::cout << *reducers[r] << "\n";
      }
      std::cout << "\n";
#endif

      std::vector< std::pair< const gcsp::clause*, size_t >> chosen;

      // Side substlets from mainparent:

      for( size_t i = mainparent -> initial. s0; 
                  i != mainparent -> initial. s1; 
                  ++ i )
      {
         // If mainparent[i] is inconsistent with the conflict
         // set of one of the lemmas, we don't add it:

         for( size_t c = 0; c < conflictlemmas. size( ); ++ c )
         {
            if( conflictlemmas[c] &&
                conflictsallconfl( *mainparent, i,
                                   *conflictlemmas[c], conflicting[c] ))
            {
               goto dontadd;
            }
         }

         if( !before_reduction. contains( i ))
         {
            // Side substlet from *mainparent:

            result. insert( *mainparent, i );
         }
         else
         {
            // Was reduced in last reduction, should be multiplied
            // with the reducers.

            chosen. push_back( { mainparent, i } );
            addproducts( result, chosen, 0, reducers );
            chosen. pop_back( ); 
         }

      dontadd:
         ;
      }

      // Side substlets from conflictlemmas:

      {
         for( size_t s = 0; s < conflictlemmas. size( ); ++ s )
         {
            if( conflictlemmas[s] )
            {
               for( size_t j = 0; 
                           j < conflictlemmas[s] -> nrsubstlets( ); ++ j )
               {
                  if( !conflicting[s]. contains(j))
                  {
                     result. insert( *conflictlemmas[s], j );
                  }
               }
            }
         }
      }

      // Side substlets from the reducers.

      for( const auto& r : reducers )
      {
         for( size_t i = r -> initial.s0; i < r -> initial.s1; ++ i )
         {
            if( ! r -> current. contains(i))
            {
               result. insert( *r, i );
            }
         }
      }

      return result;
   }


   namespace
   {

      // Count the number of variables that occur in the domain
      // of bl, but not in the domain of one the clauses in cover.


      size_t count_missing( const std::vector< const gcsp::clause* > & cover,
                            const gcsp::clause& bl )
      { 
         size_t nr = 0;

         for( size_t i = 0; i < bl. nrvars( ); ++ i )
         {
            variable v = bl. var(i);
         
            bool found = false;
            for( auto p = cover. begin( ); 
                 p != cover. end( ) && !found; ++ p )
            {
               if( (*p) -> indomain(v) )
                  found = true;
            }

            if( !found )
               ++ nr;
         }
         return nr;
      }
   }


   void add_sigma_resolvent( lemma& result, const gcsp& g, 
                             const gcsp::clause& bl, size_t j )
   {
      // std::cout << "adding sigma-resolvent for " << bl << " / " << j << "\n";

      // We start by collecting clauses that cover the variables: 

      std::vector< const gcsp::clause* > cover;

      size_t m = count_missing( cover, bl );  
      while( m )
      {
         // We greedily look for the clause that covers the highest number
         // of missing variables. If there are different clauses covering 
         // the same number, we pick the shortest clause. 

         size_t best_m = m;          // Lower is better.
         size_t best_size = 0;       // Lower is better. 
         size_t best_c = g. clauses. size( ); 

         for( size_t c = 0; c < g. clauses. size( ); ++ c ) 
         {  
            cover. push_back( &g. clauses[c] );
 
            size_t c_m = count_missing( cover, bl );

            // Fewer missing, or equal number missing but shorter clause: 

            if( c_m < best_m || 
                ( c_m == best_m && 
                  g. clauses[c]. initial. size( ) < best_size ))
            {
               best_m = c_m;
               best_size = g. clauses[c]. initial. size( );
               best_c = c; 
            } 

            cover. pop_back( );
         }

         // Because the gcsp is range restricted, we are guaranteed to
         // find something. 
  
         if( best_c == g. clauses. size( ))
            throw std::runtime_error( "gcsp not range restricted" ); 

         cover. push_back( &g. clauses[ best_c ] );
         m = count_missing( cover, bl ); 
      }

#if 0
      std::cout << "constructed the following cover:\n";
      for( const auto& p : cover )
         std::cout << *p << "\n";
      std::cout << "\n";
#endif

      // We know now which substlets will resolve away.
      // These are of course the substlets that are consistent with the
      // blocking, because all of them combined will imply the blocking.
      // The others go into the result.

      for( const gcsp::clause* p : cover )
      {
         for( size_t i = p -> initial.s0; i < p -> initial. s1; ++ i )
         {
            if( !consistent( *p, i, bl, j ))
            {
               result. insert( *p, i );
            }
         }
      }
   }   


   lemma sigma_resolvent( subststack< size_t > & theta,
                          const gcsp& g, const lemma& lem )
   {
#if 0 
      std::cout << "constructing sigma-resolvent of " << lem << "\n";
      std::cout << theta << "\n";
#endif

      lemma res;

      size_t s = theta. size( );

      for( size_t i = 0; i < lem. nrsubstlets( ); ++ i )
      {
         if( match( theta, lem, i ))
         {
            auto bl = g. getblocked( theta, s );
            if( bl. first == g. blockings. size( ))
            {
               res. insert( lem, i );
            }
            else
            {
               add_sigma_resolvent( res, g, 
                                    g. blockings[ bl. first ], bl. second );
            }
         }
         else
            res. insert( lem, i );

         theta. restore(s);
      }

      return res;
   }


}}


