
// Written by Hans de Nivelle, April 2016.
// Lemma/clause resolution, as described in the IJCAR 2016 paper.

#ifndef GEOMETRIC_CONSTRAINTS_RESOLUTION_INCLUDED
#define GEOMETRIC_CONSTRAINTS_RESOLUTION_INCLUDED  1

#include <vector>
#include <iostream>

#include "subststack.h"
#include "lemma.h"


namespace geometric 
{
namespace constraints 
{

   bool solvable( subststack< size_t > & theta,
                  const gcsp& g,
                  const gcsp::clause& cl, size_t i,
                  const gcsp::clause& cl1 );

      // True if cl/i is locally solvable against cl, 
      // taking into account the substitution theta, and the blockings in g.
      // theta will be extended and restored.

   bool solvable( subststack< size_t > & theta,
                  const gcsp& g,
                  const gcsp::clause& cl, size_t i,
                  const gcsp::clause& cl1, const gcsp::clause& cl2 );

      // True if cl/i is locally solvable against cl1, cl2,
      // taking into account the substitution theta and the blockings in g. 
      // theta will be extended and restored.

   bool implies( const gcsp& g, 
                 const std::vector< const gcsp::clause* > & clauses,
                 const std::vector< const lemma* > & lemmas,
                 const lemma& conclusion );


   lemma resolvent( const gcsp::clause* mainparent, 
                    const std::vector< const lemma* > & conflictlemmas,
                    gcsp::clause::state before_reduction,
                    const std::vector< const gcsp::clause* > & reducers );

      // mainparent -> current. size( ) must be equal to 
      // conflictlemmas. size( ). I THINK THAT THE INTERFACE SHOULD CHANGE.
      // DIFFERENT SUBSTLETS IN CURRENT MUST HAVE THE SAME CONFLICT LEMMA.
      // THERE ARE DIFFERENT WAYS TO DO THIS. MAYBE THE EASIEST WAY IS 
      // WHEN EMPTY CONFLICTLEMMAS ARE EQUAL TO THE LAST NON-EMPTY ONE.
      // Elements of conflictlemmas may be nullptr. Conflictlemmas that 
      // are present, are resolved away against their correponding substlet 
      // in mainparent.  
      // 
      // If there are substlets in mainparent -> current\before_reduction, 
      // these are multiplied with the current substlets in the reducers.
      // I BELIEVE THAT BEFORE_REDUCTION IS NOT NEEDED, IF MAINPARENT
      // IS IN PROPER STATE.
      // THE RULE MUST BE MERGED WITH SIGMA RESOLUTION. 

   void 
   add_sigma_resolvent( lemma& result, const gcsp& g, 
                        const gcsp::clause& bl, size_t j );
      // Insert the sigma-resolvent obtained from the blocking bl/j into
      // result.

   lemma sigma_resolvent( subststack< size_t > & theta, 
                          const gcsp& g, const lemma& lem ); 
      // Every substlet in lem that together with theta implies a blocking,
      // is replaced by the sigma-resolvent based on this blocking.
      // We assume that theta alone does not yet imply a blocking.

}}

#endif

