
#ifndef GEOMETRIC_PARSER_PARSER_INCLUDED
#define GEOMETRIC_PARSER_PARSER_INCLUDED  1

#include "tokenizer.h"
#include "../rulesystem.h"
#include "../interpretation.h"


namespace geometric {
namespace parser {


variable makevariable( const std::vector< std::string > & ident,
                       const ::parser::inputstream& input );

element makeelement( const std::vector< std::string > & ident, 
                     const ::parser::inputstream& input ); 

void 
advance_through( token::type t, token& lookahead, ::parser::inputstream& input );
   // Require that lookahead is a t. 
   // Throw a syntaxerror if lookahead is not a t.
   // Advance one token if lookahead is a t. 

// We maintain a general pattern, where a function is called
// with a lookahead, processes this lookahead, and leaves a new
// lookahead.

std::vector< std::string > 
   readidentifier( token& lookahead, ::parser::inputstream& input );


set< size_t > 
readassumptionset( token& lookahead, ::parser::inputstream& input );

truthval readtruthval( token& lookahead, ::parser::inputstream& input ); 

std::pair< std::vector< std::string >, size_t >
   readpredicatedeclaration( token& lookahead, ::parser::inputstream& input );

varatom readvaratom( const signature& sig,
                     token& lookahead, ::parser::inputstream& input );
groundatom readgroundatom( const signature& sig,
                           token& lookahead, ::parser::inputstream& input );

std::vector< varatom > 
readvaratomvector( const signature& sig,
                   token& lookahead, ::parser::inputstream& input );
   // Either FALSE, (epsilon), or a sequence of varatoms separated
   // by commas.

rule readrule( const signature& sig,
               token& lookahead, ::parser::inputstream& input );

std::pair< interpretation, rulesystem >
                 readproblem( signature& sig, ::parser::inputstream& input );
   
}}

#endif

