
// Written by Hans de Nivelle, December 2015.

#ifndef GEOMETRIC_PARSER_TOKEN_INCLUDED
#define GEOMETRIC_PARSER_TOKEN_INCLUDED  1

#include <string>
#include <iostream>

namespace geometric
{

   namespace parser
   {

      struct token 
      {
      
         enum class type 
         {
            ident,             //   Simple identifer. (without ::)
            number,            //   Unsigned integer. 

            separator,         //   :: 
         
            leftround,         //   ( 
            rightround,        //   )
            leftcurly,         //   {
            rightcurly,        //   }
            leftsquare,        //   [
            rightsquare,       //   ]
 
            equals,            //   ==
            create,            //   #
            comma,             //   ,
            bar,               //   |
            colon,             //   : 
            semicolon,         //   ;
            plus,              //   +
            minus,             //   - 
            slash,             //   /
            dot,               //   .

            quotedstring,      //   'bla bla ...'

            name,              //   %name
            signature,         //   %signature 
            rulesystem,        //   %rulesystem
            simplifiers,       //   %simplifiers,
            interpretation,    //   %interpretation 
            
            eof                //   Real EOF or the string %eof
         };

         type t; 
         long unsigned int i;
         std::string s;

         token( type t ) 
            : t{t},
              i{0}
         { }

         token( type t, long unsigned int i )
            : t{t},
              i{i} 
         { }

         token( type t, const std::string& s ) 
            : t{t},
              i{0},
              s{s}
         { }

      };

      
      std::ostream& operator << ( std::ostream& , token::type );
      std::ostream& operator << ( std::ostream& , const token& ); 
   }
}


#endif


