
// this file defines syntaxerror and readerror. A readerror 
// means that reading of input was attempted and failed, for some
// reason different than encountering end of file.
//
// A syntaxerror means that it was attempted to read a term, 
// that enough characters were successfully read, but that they could not
// be parsed according to the input grammar.
//
// Code written by Hans de Nivelle, April 2006.


#ifndef PARSER_ERRORS_DECLARED
#define PARSER_ERRORS_DECLARED    1 

#include "../tostring.h" 
#include <iostream>
#include "inputstream.h" 


namespace parser
{

   // A readerror is thrown when there is a problem with the inputstream,
   // and the parser cannot read from it.

   struct readerror
   { 
      std::string problem;
      std::string position; 

      // The description of the problem should be such that it can
      // be followed by 'in line ... of file ...'

      readerror( const std::string& problem, 
                 const inputstream& stream ) : 
         problem{ problem }, 
         position{ stream. inlineoffile( ) }
      {
      }

   };

   std::ostream& operator << ( std::ostream&, const readerror& );


   struct syntaxerror
   {
      std::string problem;
      std::string position;

      // The description of the problem should be such that it can
      // be followed by 'in line ... of file ...'

      syntaxerror( const std::string& problem, 
                   const inputstream& stream ) : 
         problem{ problem }, 
         position{ stream. inlineoffile( ) }
      {
      }

   };

   std::ostream& operator << ( std::ostream&, const syntaxerror& ); 
 

   struct conversionerror
   {
      std::string problem;
      std::string position;

      conversionerror( const std::string& problem )
         : problem{ problem }
      { }

      conversionerror( const std::string& problem, 
                       const inputstream& stream )
         : problem{ problem }, 
           position{ stream. inlineoffile( ) }
      {
      }

   };

   std::ostream& operator << ( std::ostream& , const conversionerror& );


} 


#endif


