
// Code written by Hans de Nivelle, April 2003. 
// Rewritten in July 2015 for CASC-25, in order to support
// includes. 

#include "inputstream.h"

#include <sstream> 

namespace parser
{

   // Whenever a stream reaches its end, it is removed from the
   // include  stack. This means that the streams on the stack supposed 
   // to be good( ).
 
   void inputstream::moveforward( ) 
   {

      if( incl. size( ))
      {
         // Do the line and column counting:

         if( incl. back( ). nextchar == '\n' )
         {
            ++ incl. back( ). linenumber;
            incl. back( ). columnnumber = 1;
         }
         else
            ++ incl. back( ). columnnumber;

         // If the file on top is not good, then something went wrong.

         if( !incl. back( ). stream. good( ))
         {
            std::ostringstream ss;
            ss << "file turned bad ";
            ss << inlineoffile( );
            throw std::runtime_error( ss. str( ));
         }

         // Read next character, this is possible because
         // the file is good( ): 
 
         incl. back( ). nextchar = incl. back( ). stream. get( );

         // If we reached end of file, we close the top file: 

         if( incl. back( ). stream. eof( ))
         {
            incl. pop_back( );
            return;  
         }
      }
      else
      {
         // Do the line and column counting:

         if( nextchar == '\n' ) 
         {
            ++ linenumber;
            columnnumber = 1;
         }
         else
            ++ columnnumber;

         if( !stream. good( ))
         {
            std::ostringstream ss;
            ss << "file turned bad ";
            ss << inlineoffile( );
            throw std::runtime_error( ss. str( ));
         }

         nextchar = stream. get( );
      }
   }

   
   void inputstream::include( const std::string& f )
   {
      incl. emplace_back(f);
      if( !incl. back( ). stream )
      {
         incl. pop_back( ); 

         std::ostringstream ss;
         ss << "could not open file " << f << " ";
         ss << inlineoffile( );
         throw std::runtime_error( ss. str( ));
      }

      moveforward( );
   }


   inputstream::~inputstream( ) 
   {
      if( nextchar != EOF ) 
         stream. putback( nextchar );
   }
 
      

   std::string inputstream::inlineoffile( ) const
   {
      std::ostringstream ss;
      ss << "in ";

      if( incl. size( ))
      {
         auto p = incl. end( );
         -- p;
         ss << "line " << ( p -> linenumber );
         ss << " of file " << ( p -> name );

         while( p != incl. begin( ))
         {
            ss << " included from ";
            -- p;
            ss << "line " << ( p -> linenumber );
            ss << " of file " << ( p -> name );
         }
 
         ss << " included from ";
      } 
      ss << "line " << linenumber;
      ss << " of file " << name;
      return ss. str( );
   }

} 


